# Copyright (c) 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import ipaddress

from oslo_log import log

from manila.common import constants
from manila.i18n import _
from manila import utils

import six

LOG = log.getLogger(__name__)


def locked_access_rules_operation(operation):
    """Lock decorator for access rules operations.

    Takes a named lock prior to executing the operation. The lock is
    named with the ID of the share instance to which the access rule belongs.

    Intended use:
    If an database operation to retrieve or update access rules uses this
    decorator, it will block actions on all access rules of the share
    instance until the named lock is free. This is used to avoid race
    conditions while performing access rules updates on a given share instance.
    """

    def wrapped(*args, **kwargs):
        instance_id = kwargs.get('share_instance_id')

        @utils.synchronized(
            "locked_access_rules_operation_by_share_instance_%s" % instance_id,
            external=True)
        def locked_operation(*_args, **_kwargs):
            return operation(*_args, **_kwargs)

        return locked_operation(*args, **kwargs)

    return wrapped


class ShareInstanceAccessDatabaseMixin(object):

    @locked_access_rules_operation
    def get_and_update_share_instance_access_rules_status(
            self, context, status=None, conditionally_change=None,
            share_instance_id=None):
        """Get and update the access_rules_status of a share instance.

        :param status: Set this parameter only if you want to
            omit the conditionally_change parameter; i.e, if you want to
            force a state change on the share instance regardless of the prior
            state.
        :param conditionally_change: Set this parameter to a dictionary of rule
            state transitions to be made. The key is the expected
            access_rules_status and the value is the state to transition the
            access_rules_status to. If the state is not as expected,
            no transition is performed. Default is {}, which means no state
            transitions will be made.
        :returns share_instance: if an update was made.
        """
        if status is not None:
            updates = {'access_rules_status': status}
        elif conditionally_change:
            share_instance = self.db.share_instance_get(
                context, share_instance_id)
            access_rules_status = share_instance['access_rules_status']
            try:
                updates = {
                    'access_rules_status':
                        conditionally_change[access_rules_status],
                }
            except KeyError:
                updates = {}
        else:
            updates = {}
        if updates:
            share_instance = self.db.share_instance_update(
                context, share_instance_id, updates, with_share_data=True)
            return share_instance

    def update_share_instances_access_rules_status(
            self, context, status, share_instance_ids):
        """Update the access_rules_status of all share instances.

        .. note::
            Before making this call, make sure that all share instances have
            their status set to a value that will block new operations to
            happen during this update.

        :param status: Force a state change on all share instances regardless
               of the prior state.
        :param share_instance_ids: List of share instance ids to have their
               access rules status updated.
        """
        updates = {'access_rules_status': status}

        self.db.share_instances_status_update(
            context, share_instance_ids, updates)

    @locked_access_rules_operation
    def get_and_update_share_instance_access_rules(self, context,
                                                   filters=None, updates=None,
                                                   conditionally_change=None,
                                                   share_instance_id=None):
        """Get and conditionally update all access rules of a share instance.

        :param updates: Set this parameter to a dictionary of key:value
            pairs corresponding to the keys in the ShareInstanceAccessMapping
            model. Include 'state' in this dictionary only if you want to
            omit the conditionally_change parameter; i.e, if you want to
            force a state change on all filtered rules regardless of the prior
            state. This parameter is always honored, regardless of whether
            conditionally_change allows for a state transition as desired.

            Example::

            {
                'access_key': 'bob007680048318f4239dfc1c192d5',
                'access_level': 'ro',
            }

        :param conditionally_change: Set this parameter to a dictionary of rule
            state transitions to be made. The key is the expected state of
            the access rule the value is the state to transition the
            access rule to. If the state is not as expected, no transition is
            performed. Default is {}, which means no state transitions
            will be made.

            Example::

            {
                'queued_to_apply': 'applying',
                'queued_to_deny': 'denying',
            }

        """
        instance_rules = self.db.share_access_get_all_for_instance(
            context, share_instance_id, filters=filters)

        if instance_rules and (updates or conditionally_change):
            if not updates:
                updates = {}
            if not conditionally_change:
                conditionally_change = {}
            for rule in instance_rules:
                mapping_state = rule['state']
                rule_updates = copy.deepcopy(updates)
                try:
                    rule_updates['state'] = conditionally_change[mapping_state]
                except KeyError:
                    pass
                if rule_updates:
                    self.db.share_instance_access_update(
                        context, rule['access_id'], share_instance_id,
                        rule_updates)

            # Refresh the rules after the updates
            rules_to_get = {
                'access_id': tuple([i['access_id'] for i in instance_rules]),
            }
            instance_rules = self.db.share_access_get_all_for_instance(
                context, share_instance_id, filters=rules_to_get)

        return instance_rules

    def get_share_instance_access_rules(self, context, filters=None,
                                        share_instance_id=None):
        return self.get_and_update_share_instance_access_rules(
            context, filters, None, None, share_instance_id)

    @locked_access_rules_operation
    def get_and_update_share_instance_access_rule(self, context, rule_id,
                                                  updates=None,
                                                  share_instance_id=None,
                                                  conditionally_change=None):
        """Get and conditionally update a given share instance access rule.

        :param updates: Set this parameter to a dictionary of key:value
            pairs corresponding to the keys in the ShareInstanceAccessMapping
            model. Include 'state' in this dictionary only if you want to
            omit the conditionally_change parameter; i.e, if you want to
            force a state change regardless of the prior state.
        :param conditionally_change: Set this parameter to a dictionary of rule
            state transitions to be made. The key is the expected state of
            the access rule the value is the state to transition the
            access rule to. If the state is not as expected, no transition is
            performed. Default is {}, which means no state transitions
            will be made.

            Example::

            {
                'queued_to_apply': 'applying',
                'queued_to_deny': 'denying',
            }
        """
        instance_rule_mapping = self.db.share_instance_access_get(
            context, rule_id, share_instance_id)

        if not updates:
            updates = {}
        if conditionally_change:
            mapping_state = instance_rule_mapping['state']
            try:
                updated_state = conditionally_change[mapping_state]
                updates.update({'state': updated_state})
            except KeyError:
                msg = ("The state of the access rule %(rule_id)s (allowing "
                       "access to share instance %(si)s) was not updated "
                       "because its state was modified by another operation.")
                msg_payload = {
                    'si': share_instance_id,
                    'rule_id': rule_id,
                }
                LOG.debug(msg, msg_payload)
        if updates:
            self.db.share_instance_access_update(
                context, rule_id, share_instance_id, updates)

            # Refresh the rule after update
            instance_rule_mapping = self.db.share_instance_access_get(
                context, rule_id, share_instance_id)

        return instance_rule_mapping

    @locked_access_rules_operation
    def delete_share_instance_access_rules(self, context, access_rules,
                                           share_instance_id=None):
        for rule in access_rules:
            self.db.share_instance_access_delete(context, rule['id'])


class ShareInstanceAccess(ShareInstanceAccessDatabaseMixin):

    def __init__(self, db, driver):
        self.db = db
        self.driver = driver

    def update_access_rules(self, context, share_instance_id,
                            delete_all_rules=False, share_server=None):
        """Update access rules for a given share instance.

        :param context: request context
        :param share_instance_id: ID of the share instance
        :param delete_all_rules: set this parameter to True if all
        existing access rules must be denied for a given share instance
        :param share_server: Share server model or None
        """
        share_instance = self.db.share_instance_get(
            context, share_instance_id, with_share_data=True)
        msg_payload = {
            'si': share_instance_id,
            'shr': share_instance['share_id'],
        }

        if delete_all_rules:
            updates = {
                'state': constants.ACCESS_STATE_QUEUED_TO_DENY,
            }
            self.get_and_update_share_instance_access_rules(
                context, updates=updates, share_instance_id=share_instance_id)

        # Is there a sync in progress? If yes, ignore the incoming request.
        rule_filter = {
            'state': (constants.ACCESS_STATE_APPLYING,
                      constants.ACCESS_STATE_DENYING),
        }
        syncing_rules = self.get_and_update_share_instance_access_rules(
            context, filters=rule_filter, share_instance_id=share_instance_id)

        if syncing_rules:
            msg = ("Access rules are being synced for share instance "
                   "%(si)s belonging to share %(shr)s, any rule changes will "
                   "be applied shortly.")
            LOG.debug(msg, msg_payload)
        else:
            rules_to_apply_or_deny = (
                self._update_and_get_unsynced_access_rules_from_db(
                    context, share_instance_id)
            )
            if rules_to_apply_or_deny:
                msg = ("Updating access rules for share instance %(si)s "
                       "belonging to share %(shr)s.")
                LOG.debug(msg, msg_payload)
                self._update_access_rules(context, share_instance_id,
                                          share_server=share_server)
            else:
                msg = ("All access rules have been synced for share instance "
                       "%(si)s belonging to share %(shr)s.")
                LOG.debug(msg, msg_payload)

    def _update_access_rules(self, context, share_instance_id,
                             share_server=None):
        # Refresh the share instance model
        share_instance = self.db.share_instance_get(
            context, share_instance_id, with_share_data=True)

        conditionally_change = {
            constants.STATUS_ACTIVE: constants.SHARE_INSTANCE_RULES_SYNCING,
        }
        share_instance = (
            self.get_and_update_share_instance_access_rules_status(
                context, conditionally_change=conditionally_change,
                share_instance_id=share_instance_id) or share_instance
        )

        rules_to_be_removed_from_db = []
        # Populate rules to send to the driver
        (access_rules_to_be_on_share, add_rules, delete_rules) = (
            self._get_rules_to_send_to_driver(context, share_instance)
        )

        if share_instance['cast_rules_to_readonly']:
            # Ensure read/only semantics for a migrating instances
            access_rules_to_be_on_share = self._set_rules_to_readonly(
                access_rules_to_be_on_share, share_instance)
            add_rules = []
            rules_to_be_removed_from_db = delete_rules
            delete_rules = []

        try:
            driver_rule_updates = self._update_rules_through_share_driver(
                context, share_instance, access_rules_to_be_on_share,
                add_rules, delete_rules, rules_to_be_removed_from_db,
                share_server)

            self.process_driver_rule_updates(
                context, driver_rule_updates, share_instance_id)

            # Update access rules that are still in 'applying' state
            conditionally_change = {
                constants.ACCESS_STATE_APPLYING: constants.ACCESS_STATE_ACTIVE,
            }
            self.get_and_update_share_instance_access_rules(
                context, share_instance_id=share_instance_id,
                conditionally_change=conditionally_change)

        except Exception:
            conditionally_change_rule_state = {
                constants.ACCESS_STATE_APPLYING: constants.ACCESS_STATE_ERROR,
                constants.ACCESS_STATE_DENYING: constants.ACCESS_STATE_ERROR,
            }
            self.get_and_update_share_instance_access_rules(
                context, share_instance_id=share_instance_id,
                conditionally_change=conditionally_change_rule_state)

            conditionally_change_access_rules_status = {
                constants.ACCESS_STATE_ACTIVE: constants.STATUS_ERROR,
                constants.SHARE_INSTANCE_RULES_SYNCING: constants.STATUS_ERROR,
            }
            self.get_and_update_share_instance_access_rules_status(
                context, share_instance_id=share_instance_id,
                conditionally_change=conditionally_change_access_rules_status)
            raise

        if rules_to_be_removed_from_db:
            delete_rules = rules_to_be_removed_from_db

        self.delete_share_instance_access_rules(
            context, delete_rules, share_instance_id=share_instance['id'])

        self._loop_for_refresh_else_update_access_rules_status(
            context, share_instance_id, share_server)

        msg = _("Access rules were successfully modified for share instance "
                "%(si)s belonging to share %(shr)s.")
        msg_payload = {
            'si': share_instance['id'],
            'shr': share_instance['share_id'],
        }
        LOG.info(msg, msg_payload)

    def _update_rules_through_share_driver(self, context, share_instance,
                                           access_rules_to_be_on_share,
                                           add_rules, delete_rules,
                                           rules_to_be_removed_from_db,
                                           share_server):
        driver_rule_updates = {}
        share_protocol = share_instance['share_proto'].lower()
        if (not self.driver.ipv6_implemented and
                share_protocol == 'nfs'):
            add_rules = self._filter_ipv6_rules(add_rules)
            delete_rules = self._filter_ipv6_rules(delete_rules)
            access_rules_to_be_on_share = self._filter_ipv6_rules(
                access_rules_to_be_on_share)
        try:
            driver_rule_updates = self.driver.update_access(
                context,
                share_instance,
                access_rules_to_be_on_share,
                add_rules=add_rules,
                delete_rules=delete_rules,
                share_server=share_server
            ) or {}
        except NotImplementedError:
            # NOTE(u_glide): Fallback to legacy allow_access/deny_access
            # for drivers without update_access() method support
            self._update_access_fallback(context, add_rules, delete_rules,
                                         rules_to_be_removed_from_db,
                                         share_instance,
                                         share_server)
        return driver_rule_updates

    def _loop_for_refresh_else_update_access_rules_status(self, context,
                                                          share_instance_id,
                                                          share_server):
        # Do we need to re-sync or apply any new changes?
        if self._check_needs_refresh(context, share_instance_id):
            self._update_access_rules(context, share_instance_id,
                                      share_server=share_server)
        else:
            # Switch the share instance's access_rules_status to 'active'
            # if there are no more rules in 'error' state, else, ensure
            # 'error' state.
            rule_filter = {'state': constants.STATUS_ERROR}
            rules_in_error_state = (
                self.get_and_update_share_instance_access_rules(
                    context, filters=rule_filter,
                    share_instance_id=share_instance_id)
            )
            if not rules_in_error_state:
                conditionally_change = {
                    constants.SHARE_INSTANCE_RULES_SYNCING:
                        constants.STATUS_ACTIVE,
                    constants.SHARE_INSTANCE_RULES_ERROR:
                        constants.STATUS_ACTIVE,
                }
                self.get_and_update_share_instance_access_rules_status(
                    context, conditionally_change=conditionally_change,
                    share_instance_id=share_instance_id)
            else:
                conditionally_change = {
                    constants.SHARE_INSTANCE_RULES_SYNCING:
                        constants.SHARE_INSTANCE_RULES_ERROR,
                }
                self.get_and_update_share_instance_access_rules_status(
                    context, conditionally_change=conditionally_change,
                    share_instance_id=share_instance_id)

    def process_driver_rule_updates(self, context, driver_rule_updates,
                                    share_instance_id):
        for rule_id, rule_updates in driver_rule_updates.items():
            if 'state' in rule_updates:
                # We allow updates *only* if the state is unchanged from
                # the time this update was initiated. It is possible
                # that the access rule was denied at the API prior to
                # the driver reporting that the access rule was added
                # successfully.
                state = rule_updates.pop('state')
                conditional_state_updates = {
                    constants.ACCESS_STATE_APPLYING: state,
                    constants.ACCESS_STATE_DENYING: state,
                    constants.ACCESS_STATE_ACTIVE: state,
                }
            else:
                conditional_state_updates = {}
            self.get_and_update_share_instance_access_rule(
                context, rule_id, updates=rule_updates,
                share_instance_id=share_instance_id,
                conditionally_change=conditional_state_updates)

    @staticmethod
    def _set_rules_to_readonly(access_rules_to_be_on_share, share_instance):

        LOG.debug("All access rules of share instance %s are being "
                  "cast to read-only for a migration or because the "
                  "instance is a readable replica.",
                  share_instance['id'])

        for rule in access_rules_to_be_on_share:
            rule['access_level'] = constants.ACCESS_LEVEL_RO

        return access_rules_to_be_on_share

    @staticmethod
    def _filter_ipv6_rules(rules):
        filtered = []
        for rule in rules:
            if rule['access_type'] == 'ip':
                ip_version = ipaddress.ip_network(
                    six.text_type(rule['access_to'])).version
                if 6 == ip_version:
                    continue
            filtered.append(rule)
        return filtered

    def _get_rules_to_send_to_driver(self, context, share_instance):
        add_rules = []
        delete_rules = []
        access_filters = {
            'state': (constants.ACCESS_STATE_APPLYING,
                      constants.ACCESS_STATE_ACTIVE,
                      constants.ACCESS_STATE_DENYING),
        }
        existing_rules_in_db = self.get_and_update_share_instance_access_rules(
            context, filters=access_filters,
            share_instance_id=share_instance['id'])
        # Update queued rules to transitional states
        for rule in existing_rules_in_db:

            if rule['state'] == constants.ACCESS_STATE_APPLYING:
                add_rules.append(rule)
            elif rule['state'] == constants.ACCESS_STATE_DENYING:
                delete_rules.append(rule)
        delete_rule_ids = [r['id'] for r in delete_rules]
        access_rules_to_be_on_share = [
            r for r in existing_rules_in_db if r['id'] not in delete_rule_ids
        ]
        return access_rules_to_be_on_share, add_rules, delete_rules

    def _check_needs_refresh(self, context, share_instance_id):
        rules_to_apply_or_deny = (
            self._update_and_get_unsynced_access_rules_from_db(
                context, share_instance_id)
        )
        return any(rules_to_apply_or_deny)

    def _update_access_fallback(self, context, add_rules, delete_rules,
                                remove_rules, share_instance, share_server):
        for rule in add_rules:
            LOG.info(
                "Applying access rule '%(rule)s' for share "
                "instance '%(instance)s'",
                {'rule': rule['id'], 'instance': share_instance['id']}
            )

            self.driver.allow_access(
                context,
                share_instance,
                rule,
                share_server=share_server
            )

        # NOTE(ganso): Fallback mode temporary compatibility workaround
        if remove_rules:
            delete_rules.extend(remove_rules)

        for rule in delete_rules:
            LOG.info(
                "Denying access rule '%(rule)s' from share "
                "instance '%(instance)s'",
                {'rule': rule['id'], 'instance': share_instance['id']}
            )

            self.driver.deny_access(
                context,
                share_instance,
                rule,
                share_server=share_server
            )

    def _update_and_get_unsynced_access_rules_from_db(self, context,
                                                      share_instance_id):
        rule_filter = {
            'state': (constants.ACCESS_STATE_QUEUED_TO_APPLY,
                      constants.ACCESS_STATE_QUEUED_TO_DENY),
        }
        conditionally_change = {
            constants.ACCESS_STATE_QUEUED_TO_APPLY:
                constants.ACCESS_STATE_APPLYING,
            constants.ACCESS_STATE_QUEUED_TO_DENY:
                constants.ACCESS_STATE_DENYING,
        }
        rules_to_apply_or_deny = (
            self.get_and_update_share_instance_access_rules(
                context, filters=rule_filter,
                share_instance_id=share_instance_id,
                conditionally_change=conditionally_change)
        )
        return rules_to_apply_or_deny

    def reset_applying_rules(self, context, share_instance_id):
        conditional_updates = {
            constants.ACCESS_STATE_APPLYING:
                constants.ACCESS_STATE_QUEUED_TO_APPLY,
            constants.ACCESS_STATE_DENYING:
                constants.ACCESS_STATE_QUEUED_TO_DENY,
        }
        self.get_and_update_share_instance_access_rules(
            context, share_instance_id=share_instance_id,
            conditionally_change=conditional_updates)



# Copyright (c) 2016 Hitachi Data Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from manila.api.openstack import wsgi
from manila.api.views import share_snapshot_export_locations
from manila.db import api as db_api
from manila import exception
from manila.i18n import _
from manila import policy


class ShareSnapshotInstanceExportLocationController(wsgi.Controller):

    def __init__(self):
        self._view_builder_class = (
            share_snapshot_export_locations.ViewBuilder)
        self.resource_name = 'share_snapshot_instance_export_location'
        super(ShareSnapshotInstanceExportLocationController, self).__init__()

    @wsgi.Controller.api_version('2.32')
    @wsgi.Controller.authorize
    def index(self, req, snapshot_instance_id):
        context = req.environ['manila.context']
        instance = self._verify_snapshot_instance(
            context, snapshot_instance_id)
        export_locations = (
            db_api.share_snapshot_instance_export_locations_get_all(
                context, instance['id']))

        return self._view_builder.list_export_locations(req, export_locations)

    @wsgi.Controller.api_version('2.32')
    @wsgi.Controller.authorize
    def show(self, req, snapshot_instance_id, export_location_id):
        context = req.environ['manila.context']
        self._verify_snapshot_instance(context, snapshot_instance_id)
        export_location = db_api.share_snapshot_instance_export_location_get(
            context, export_location_id)
        return self._view_builder.detail_export_location(req, export_location)

    def _verify_snapshot_instance(self, context, snapshot_instance_id):
        try:
            snapshot_instance = db_api.share_snapshot_instance_get(
                context, snapshot_instance_id)
            share = db_api.share_get(
                context, snapshot_instance.share_instance['share_id'])
            if not share['is_public']:
                policy.check_policy(context, 'share', 'get', share)
        except exception.NotFound:
            msg = _("Snapshot instance '%s' not found.") % snapshot_instance_id
            raise exc.HTTPNotFound(explanation=msg)
        return snapshot_instance


def create_resource():
    return wsgi.Resource(ShareSnapshotInstanceExportLocationController())



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add 'snapshot_support' extra spec to share types

Revision ID: 55761e5f59c5
Revises: 1f0bd302c1a6
Create Date: 2015-08-13 14:02:54.656864

"""

# revision identifiers, used by Alembic.
revision = '55761e5f59c5'
down_revision = '1f0bd302c1a6'

from alembic import op
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy.sql import table

from manila.common import constants


def upgrade():
    """Performs DB upgrade to support feature of making snapshots optional.

    Add 'snapshot_support' extra spec to all share types and
    attr 'snapshot_support' to Share model.
    """
    session = sa.orm.Session(bind=op.get_bind().connect())

    es_table = table(
        'share_type_extra_specs',
        sa.Column('created_at', sa.DateTime),
        sa.Column('deleted', sa.Integer),
        sa.Column('share_type_id', sa.String(length=36)),
        sa.Column('spec_key', sa.String(length=255)),
        sa.Column('spec_value', sa.String(length=255)))

    st_table = table(
        'share_types',
        sa.Column('deleted', sa.Integer),
        sa.Column('id', sa.Integer))

    # NOTE(vponomaryov): field 'deleted' is integer here.
    existing_extra_specs = (session.query(es_table).
                            filter(es_table.c.spec_key ==
                                   constants.ExtraSpecs.SNAPSHOT_SUPPORT).
                            filter(es_table.c.deleted == 0).
                            all())
    exclude_st_ids = [es.share_type_id for es in existing_extra_specs]

    # NOTE(vponomaryov): field 'deleted' is string here.
    share_types = (session.query(st_table).
                   filter(st_table.c.deleted.in_(('0', 'False', ))).
                   filter(st_table.c.id.notin_(exclude_st_ids)).
                   all())
    session.close_all()

    extra_specs = []
    now = timeutils.utcnow()
    for st in share_types:
        extra_specs.append({
            'spec_key': constants.ExtraSpecs.SNAPSHOT_SUPPORT,
            'spec_value': 'True',
            'deleted': 0,
            'created_at': now,
            'share_type_id': st.id,
        })
    if extra_specs:
        op.bulk_insert(es_table, extra_specs)

    # NOTE(vponomaryov): shares that were created before applying this
    # migration can have incorrect value because they were created without
    # consideration of driver capability to create snapshots.
    op.add_column('shares',
                  sa.Column('snapshot_support', sa.Boolean, default=True))

    connection = op.get_bind().connect()
    shares = sa.Table(
        'shares',
        sa.MetaData(),
        autoload=True,
        autoload_with=connection)

    # pylint: disable=no-value-for-parameter
    update = shares.update().where(shares.c.deleted == 'False').values(
        snapshot_support=True)
    connection.execute(update)


def downgrade():
    """Performs DB downgrade removing support of 'optional snapshots' feature.

    Remove 'snapshot_support' extra spec from all share types and
    attr 'snapshot_support' from Share model.
    """
    connection = op.get_bind().connect()
    extra_specs = sa.Table(
        'share_type_extra_specs',
        sa.MetaData(),
        autoload=True,
        autoload_with=connection)

    # pylint: disable=no-value-for-parameter
    update = extra_specs.update().where(
        extra_specs.c.spec_key == constants.ExtraSpecs.SNAPSHOT_SUPPORT).where(
        extra_specs.c.deleted == 0).values(
            deleted=extra_specs.c.id,
            deleted_at=timeutils.utcnow(),
    )
    connection.execute(update)

    op.drop_column('shares', 'snapshot_support')



# Copyright 2019 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Driver for Inspur InStorage
"""

import ipaddress
import itertools
import six

from oslo_config import cfg
from oslo_log import log
from oslo_utils import units

from manila import coordination
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share import utils as share_utils

from manila.share.drivers.inspur.instorage.cli_helper import InStorageSSH
from manila.share.drivers.inspur.instorage.cli_helper import SSHRunner

instorage_opts = [
    cfg.HostAddressOpt(
        'instorage_nas_ip',
        required=True,
        help='IP address for the InStorage.'
    ),
    cfg.PortOpt(
        'instorage_nas_port',
        default=22,
        help='Port number for the InStorage.'
    ),
    cfg.StrOpt(
        'instorage_nas_login',
        required=True,
        help='Username for the InStorage.'
    ),
    cfg.StrOpt(
        'instorage_nas_password',
        required=True,
        secret=True,
        help='Password for the InStorage.'
    ),
    cfg.ListOpt(
        'instorage_nas_pools',
        required=True,
        help='The Storage Pools Manila should use, a comma separated list.'
    )
]

CONF = cfg.CONF
CONF.register_opts(instorage_opts)
LOG = log.getLogger(__name__)


class InStorageShareDriver(driver.ShareDriver):
    """Inspur InStorage NAS driver. Allows for NFS and CIFS NAS.

    .. code::none
        Version history:
            1.0.0 - Initial driver.
                    Driver support:
                        share create/delete
                        extend size
                        update_access
                        protocol: NFS/CIFS
    """

    VENDOR = 'INSPUR'
    VERSION = '1.0.0'
    PROTOCOL = 'NFS_CIFS'

    def __init__(self, *args, **kwargs):
        super(InStorageShareDriver, self).__init__(False, *args, **kwargs)
        self.configuration.append_config_values(instorage_opts)

        self.backend_name = self.configuration.safe_get('share_backend_name')
        self.backend_pools = self.configuration.instorage_nas_pools

        self.ssh_runner = SSHRunner(**{
            'host': self.configuration.instorage_nas_ip,
            'port': 22,
            'login': self.configuration.instorage_nas_login,
            'password': self.configuration.instorage_nas_password
        })

        self.assistant = InStorageAssistant(self.ssh_runner)

    def check_for_setup_error(self):
        nodes = self.assistant.get_nodes_info()
        if len(nodes) == 0:
            msg = _('No valid node, be sure the NAS Port IP is configured')
            raise exception.ShareBackendException(msg=msg)

        pools = self.assistant.get_available_pools()
        not_exist = set(self.backend_pools).difference(set(pools))
        if not_exist:
            msg = _('Pool %s not exist on the storage system') % not_exist
            raise exception.InvalidParameterValue(msg)

    def _update_share_stats(self, **kwargs):
        """Retrieve share stats information."""

        try:
            stats = {
                'share_backend_name': self.backend_name,
                'vendor_name': self.VENDOR,
                'driver_version': self.VERSION,
                'storage_protocol': 'NFS_CIFS',
                'reserved_percentage':
                    self.configuration.reserved_share_percentage,
                'reserved_snapshot_percentage': (
                    self.configuration.reserved_share_from_snapshot_percentage
                    or self.configuration.reserved_share_percentage),
                'max_over_subscription_ratio':
                    self.configuration.max_over_subscription_ratio,
                'snapshot_support': False,
                'create_share_from_snapshot_support': False,
                'revert_to_snapshot_support': False,
                'qos': False,
                'total_capacity_gb': 0.0,
                'free_capacity_gb': 0.0,
                'pools': []
            }

            pools = self.assistant.get_pools_attr(self.backend_pools)
            total_capacity_gb = 0
            free_capacity_gb = 0
            for pool in pools.values():
                total_capacity_gb += pool['total_capacity_gb']
                free_capacity_gb += pool['free_capacity_gb']
                stats['pools'].append(pool)

            stats['total_capacity_gb'] = total_capacity_gb
            stats['free_capacity_gb'] = free_capacity_gb

            LOG.debug('share status %s', stats)

            super(InStorageShareDriver, self)._update_share_stats(stats)
        except Exception:
            msg = _('Unexpected error while trying to get the '
                    'usage stats from array.')
            LOG.exception(msg)
            raise

    @staticmethod
    def generate_share_name(share):
        # Generate a name with id of the share as base, and do follows:
        # 1. Remove the '-' in the id string.
        # 2. Transform all alpha to lower case.
        # 3. If the first char of the id is a num,
        #    transform it to an Upper case alpha start from 'A',
        #    such as '0' -> 'A', '1' -> 'B'.
        # e.g.
        # generate_share_name({
        #  'id': '46CF5E85-D618-4023-8727-6A1EA9292954',
        #  ...
        # })
        # returns 'E6cf5e85d618402387276a1ea9292954'

        name = share['id'].replace('-', '').lower()
        if name[0] in '0123456789':
            name = chr(ord('A') + ord(name[0]) - ord('0')) + name[1:]
        return name

    def get_network_allocations_number(self):
        """Get the number of network interfaces to be created."""

        return 0

    def create_share(self, context, share, share_server=None):
        """Create a new share instance."""
        share_name = self.generate_share_name(share)
        share_size = share['size']
        share_proto = share['share_proto']

        pool_name = share_utils.extract_host(share['host'], level='pool')

        self.assistant.create_share(
            share_name,
            pool_name,
            share_size,
            share_proto
        )

        return self.assistant.get_export_locations(share_name, share_proto)

    def delete_share(self, context, share, share_server=None):
        """Delete the given share instance."""
        share_name = self.generate_share_name(share)
        share_proto = share['share_proto']

        self.assistant.delete_share(share_name, share_proto)

    def extend_share(self, share, new_size, share_server=None):
        """Extend the share instance's size to new size."""
        share_name = self.generate_share_name(share)

        self.assistant.extend_share(share_name, new_size)

    def ensure_share(self, context, share, share_server=None):
        """Ensure that the share instance is exported."""
        share_name = self.generate_share_name(share)
        share_proto = share['share_proto']

        return self.assistant.get_export_locations(share_name, share_proto)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update the share instance's access rule."""
        share_name = self.generate_share_name(share)
        share_proto = share['share_proto']

        @coordination.synchronized('inspur-instorage-access-' + share_name)
        def _update_access(name, proto, rules, add_rules, delete_rules):
            self.assistant.update_access(
                name, proto, rules, add_rules, delete_rules
            )

        _update_access(
            share_name, share_proto, access_rules, add_rules, delete_rules
        )


class InStorageAssistant(object):

    NFS_CLIENT_SPEC_PATTERN = (
        '%(ip)s/%(mask)s:%(rights)s:%(all_squash)s:%(root_squash)s'
    )

    CIFS_CLIENT_RIGHT_PATTERN = (
        '%(type)s:%(name)s:%(rights)s'
    )

    def __init__(self, ssh_runner):
        self.ssh = InStorageSSH(ssh_runner)

    @staticmethod
    def handle_keyerror(cmd, out):
        msg = (_('Could not find key in output of command %(cmd)s: %(out)s.')
               % {'out': out, 'cmd': cmd})
        raise exception.ShareBackendException(msg=msg)

    def size_to_gb(self, size):
        new_size = 0

        if 'P' in size:
            new_size = int(float(size.rstrip('PB')) * units.Mi)
        elif 'T' in size:
            new_size = int(float(size.rstrip('TB')) * units.Ki)
        elif 'G' in size:
            new_size = int(float(size.rstrip('GB')) * 1)
        elif 'M' in size:
            mb_size = float(size.rstrip('MB'))
            new_size = int((mb_size + units.Ki - 1) / units.Ki)

        return new_size

    def get_available_pools(self):
        nas_pools = self.ssh.lsnaspool()
        return [pool['pool_name'] for pool in nas_pools]

    def get_pools_attr(self, backend_pools):
        pools = {}
        fs_attr = self.ssh.lsfs()
        nas_pools = self.ssh.lsnaspool()
        for pool_attr in nas_pools:
            pool_name = pool_attr['pool_name']
            if pool_name not in backend_pools:
                continue

            total_used_capacity = 0
            total_allocated_capacity = 0
            for fs in fs_attr:
                if fs['pool_name'] != pool_name:
                    continue
                allocated = self.size_to_gb(fs['total_capacity'])
                used = self.size_to_gb(fs['used_capacity'])

                total_allocated_capacity += allocated
                total_used_capacity += used

            available = self.size_to_gb(pool_attr['available_capacity'])

            pool = {
                'pool_name': pool_name,
                'total_capacity_gb': total_allocated_capacity + available,
                'free_capacity_gb': available,
                'allocated_capacity_gb': total_allocated_capacity,
                'reserved_percentage': 0,
                'reserved_snapshot_percentage': 0,
                'qos': False,
                'dedupe': False,
                'compression': False,
                'thin_provisioning': False,
                'max_over_subscription_ratio': 0
            }

            pools[pool_name] = pool

        return pools

    def get_nodes_info(self):
        """Return a dictionary containing information of system's nodes."""
        nodes = {}
        resp = self.ssh.lsnasportip()
        for port in resp:
            try:
                # Port is invalid if it has no IP configured.
                if port['ip'] == '':
                    continue

                node_name = port['node_name']
                if node_name not in nodes:
                    nodes[node_name] = {}

                node = nodes[node_name]
                node[port['id']] = port
            except KeyError:
                self.handle_keyerror('lsnasportip', port)

        return nodes

    @staticmethod
    def get_fsname_by_name(name):
        return ('%(fsname)s' % {'fsname': name})[0:32]

    @staticmethod
    def get_dirname_by_name(name):
        return ('%(dirname)s' % {'dirname': name})[0:32]

    def get_dirpath_by_name(self, name):
        fsname = self.get_fsname_by_name(name)
        dirname = self.get_dirname_by_name(name)

        return '/fs/%(fsname)s/%(dirname)s' % {
            'fsname': fsname, 'dirname': dirname
        }

    def create_share(self, name, pool, size, proto):
        """Create a share with given info."""

        # use one available node as the primary node
        nodes = self.get_nodes_info()
        if len(nodes) == 0:
            msg = _('No valid node, be sure the NAS Port IP is configured')
            raise exception.ShareBackendException(msg=msg)

        node_name = [key for key in nodes.keys()][0]

        # first create the file system on which share will be created
        fsname = self.get_fsname_by_name(name)
        self.ssh.addfs(fsname, pool, size, node_name)

        # then create the directory used for the share
        dirpath = self.get_dirpath_by_name(name)
        self.ssh.addnasdir(dirpath)

        # For CIFS, we need to create a CIFS share.
        # For NAS, the share is automatically added when the first
        # 'access spec' is added on it.
        if proto == 'CIFS':
            self.ssh.addcifs(name, dirpath)

    def check_share_exist(self, name):
        """Check whether the specified share exist on backend."""

        fsname = self.get_fsname_by_name(name)
        for fs in self.ssh.lsfs():
            if fs['fs_name'] == fsname:
                return True
        return False

    def delete_share(self, name, proto):
        """Delete the given share."""

        if not self.check_share_exist(name):
            LOG.warning('Share %s does not exist on the backend.', name)
            return

        # For CIFS, we have to delete the share first.
        # For NAS, when the last client access spec is removed from
        # it, the share is automatically deleted.
        if proto == 'CIFS':
            self.ssh.rmcifs(name)

        # then delete the directory
        dirpath = self.get_dirpath_by_name(name)
        self.ssh.rmnasdir(dirpath)

        # at last delete the file system
        fsname = self.get_fsname_by_name(name)
        self.ssh.rmfs(fsname)

    def extend_share(self, name, new_size):
        """Extend a given share to a new size.

        :param name: the name of the share.
        :param new_size: the new size the share should be.
        :return:
        """
        # first get the original capacity
        old_size = None
        fsname = self.get_fsname_by_name(name)
        for fs in self.ssh.lsfs():
            if fs['fs_name'] == fsname:
                old_size = self.size_to_gb(fs['total_capacity'])
                break

        if old_size is None:
            msg = _('share %s is not available') % name
            raise exception.ShareBackendException(msg=msg)

        LOG.debug('Extend fs %s from %dGB to %dGB', fsname, old_size, new_size)
        self.ssh.expandfs(fsname, new_size - old_size)

    def get_export_locations(self, name, share_proto):
        """Get the export locations of a given share.

        :param name: the name of the share.
        :param share_proto: the protocol of the share.
        :return: a list of export locations.
        """

        if share_proto == 'NFS':
            dirpath = self.get_dirpath_by_name(name)
            pattern = '%(ip)s:' + dirpath
        elif share_proto == 'CIFS':
            pattern = '\\\\%(ip)s\\' + name
        else:
            msg = _('share protocol %s is not supported') % share_proto
            raise exception.ShareBackendException(msg=msg)

        # we need get the node so that we know which port ip we can use
        node_name = None
        fsname = self.get_fsname_by_name(name)
        for node in self.ssh.lsnode():
            for fs in self.ssh.lsfs(node['name']):
                if fs['fs_name'] == fsname:
                    node_name = node['name']
                    break
            if node_name:
                break

        if node_name is None:
            msg = _('share %s is not available') % name
            raise exception.ShareBackendException(msg=msg)

        locations = []
        ports = self.ssh.lsnasportip()
        for port in ports:
            if port['node_name'] == node_name and port['ip'] != '':
                location = pattern % {'ip': port['ip']}

                locations.append({
                    'path': location,
                    'is_admin_only': False,
                    'metadata': {}
                })

        return locations

    def classify_nfs_client_spec(self, client_spec, dirpath):
        nfslist = self.ssh.lsnfslist(dirpath)
        if len(nfslist):
            nfsinfo = self.ssh.lsnfsinfo(dirpath)
            spec_set = set([
                self.NFS_CLIENT_SPEC_PATTERN % i for i in nfsinfo
            ])
        else:
            spec_set = set()

        client_spec_set = set(client_spec)

        del_spec = spec_set.difference(client_spec_set)
        add_spec = client_spec_set.difference(spec_set)

        return list(add_spec), list(del_spec)

    def access_rule_to_client_spec(self, access_rule):
        if access_rule['access_type'] != 'ip':
            msg = _('only ip access type is supported when using NFS protocol')
            raise exception.ShareBackendException(msg=msg)

        network = ipaddress.ip_network(six.text_type(access_rule['access_to']))
        if network.version != 4:
            msg = _('only IPV4 is accepted when using NFS protocol')
            raise exception.ShareBackendException(msg=msg)

        client_spec = self.NFS_CLIENT_SPEC_PATTERN % {
            'ip': six.text_type(network.network_address),
            'mask': six.text_type(network.netmask),
            'rights': access_rule['access_level'],
            'all_squash': 'all_squash',
            'root_squash': 'root_squash'
        }

        return client_spec

    def update_nfs_access(self, share_name, access_rules, add_rules,
                          delete_rules):
        """Update a NFS share's access rule."""

        dirpath = self.get_dirpath_by_name(share_name)
        if add_rules or delete_rules:
            add_spec = [
                self.access_rule_to_client_spec(r) for r in add_rules
            ]
            del_spec = [
                self.access_rule_to_client_spec(r) for r in delete_rules
            ]

            _, can_del_spec = self.classify_nfs_client_spec(
                [], dirpath
            )
            to_del_set = set(del_spec)
            can_del_set = set(can_del_spec)
            will_del_set = to_del_set.intersection(can_del_set)
            del_spec = list(will_del_set)
        else:
            access_spec = [
                self.access_rule_to_client_spec(r) for r in access_rules
            ]

            add_spec, del_spec = self.classify_nfs_client_spec(
                access_spec, dirpath
            )

        for spec in del_spec:
            self.ssh.rmnfsclient(dirpath, spec)
        for spec in add_spec:
            self.ssh.addnfsclient(dirpath, spec)

    def classify_cifs_rights(self, access_rights, share_name):
        cifsinfo = self.ssh.lscifsinfo(share_name)
        rights_set = set([
            self.CIFS_CLIENT_RIGHT_PATTERN % i for i in cifsinfo
        ])
        access_rights_set = set(access_rights)

        del_rights = rights_set.difference(access_rights_set)
        add_rights = access_rights_set.difference(rights_set)

        return list(add_rights), list(del_rights)

    def access_rule_to_rights(self, access_rule):
        if access_rule['access_type'] != 'user':
            msg = _('only user access type is supported'
                    ' when using CIFS protocol')
            raise exception.ShareBackendException(msg=msg)

        rights = self.CIFS_CLIENT_RIGHT_PATTERN % {
            'type': 'LU',
            'name': access_rule['access_to'],
            'rights': access_rule['access_level']
        }

        return rights

    def update_cifs_access(self, share_name, access_rules, add_rules,
                           delete_rules):
        """Update a CIFS share's access rule."""

        if add_rules or delete_rules:
            add_rights = [
                self.access_rule_to_rights(r) for r in add_rules
            ]
            del_rights = [
                self.access_rule_to_rights(r) for r in delete_rules
            ]
        else:
            access_rights = [
                self.access_rule_to_rights(r) for r in access_rules
            ]

            add_rights, del_rights = self.classify_cifs_rights(
                access_rights, share_name
            )

        for rights in del_rights:
            self.ssh.rmcifsuser(share_name, rights)
        for rights in add_rights:
            self.ssh.addcifsuser(share_name, rights)

    @staticmethod
    def check_access_type(access_type, *rules):
        rule_chain = itertools.chain(*rules)
        if all([r['access_type'] == access_type for r in rule_chain]):
            return True
        else:
            return False

    def update_access(self, share_name, share_proto,
                      access_rules, add_rules, delete_rules):
        if share_proto == 'CIFS':
            if self.check_access_type('user', access_rules,
                                      add_rules, delete_rules):
                self.update_cifs_access(share_name, access_rules,
                                        add_rules, delete_rules)
            else:
                msg = _("Only %s access type allowed.") % "user"
                raise exception.InvalidShareAccess(reason=msg)
        elif share_proto == 'NFS':
            if self.check_access_type('ip', access_rules,
                                      add_rules, delete_rules):
                self.update_nfs_access(share_name, access_rules,
                                       add_rules, delete_rules)
            else:
                msg = _("Only %s access type allowed.") % "ip"
                raise exception.InvalidShareAccess(reason=msg)
        else:
            msg = _('share protocol %s is not supported') % share_proto
            raise exception.ShareBackendException(msg=msg)



# Copyright (c) 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils
from oslo_utils import strutils
import six

from manila import exception
from manila.i18n import _
from manila import utils

LOG = log.getLogger(__name__)
CONF = cfg.CONF

winrm_opts = [
    cfg.IntOpt(
        'winrm_conn_timeout',
        default=60,
        help='WinRM connection timeout.'),
    cfg.IntOpt(
        'winrm_operation_timeout',
        default=60,
        help='WinRM operation timeout.'),
    cfg.IntOpt(
        'winrm_retry_count',
        default=3,
        help='WinRM retry count.'),
    cfg.IntOpt(
        'winrm_retry_interval',
        default=5,
        help='WinRM retry interval in seconds'),
]

CONF.register_opts(winrm_opts)

DEFAULT_PORT_HTTP = 5985
DEFAULT_PORT_HTTPS = 5986

TRANSPORT_PLAINTEXT = 'plaintext'
TRANSPORT_SSL = 'ssl'

winrm = None


def setup_winrm():
    global winrm
    if not winrm:
        try:
            winrm = importutils.import_module('winrm')
        except ImportError:
            raise exception.ShareBackendException(
                _("PyWinrm is not installed"))


class WinRMHelper(object):
    def __init__(self, configuration=None):
        if configuration:
            configuration.append_config_values(winrm_opts)
            self._config = configuration
        else:
            self._config = CONF

        setup_winrm()

    def _get_conn(self, server):
        auth = self._get_auth(server)
        conn = WinRMConnection(
            ip=server['ip'],
            conn_timeout=self._config.winrm_conn_timeout,
            operation_timeout=self._config.winrm_operation_timeout,
            **auth)
        return conn

    def execute(self, server, command, check_exit_code=True,
                retry=True):
        retries = self._config.winrm_retry_count if retry else 1
        conn = self._get_conn(server)

        @utils.retry(retry_param=Exception,
                     interval=self._config.winrm_retry_interval,
                     retries=retries)
        def _execute():
            parsed_cmd, sanitized_cmd = self._parse_command(command)

            LOG.debug("Executing command: %s", sanitized_cmd)
            (stdout, stderr, exit_code) = conn.execute(parsed_cmd)

            sanitized_stdout = strutils.mask_password(stdout)
            sanitized_stderr = strutils.mask_password(stderr)
            LOG.debug("Executed command: %(cmd)s. Stdout: %(stdout)s. "
                      "Stderr: %(stderr)s. Exit code %(exit_code)s",
                      dict(cmd=sanitized_cmd, stdout=sanitized_stdout,
                           stderr=sanitized_stderr, exit_code=exit_code))

            if check_exit_code and exit_code != 0:
                raise processutils.ProcessExecutionError(
                    stdout=sanitized_stdout,
                    stderr=sanitized_stderr,
                    exit_code=exit_code,
                    cmd=sanitized_cmd)
            return (stdout, stderr)
        return _execute()

    def _parse_command(self, command):
        if isinstance(command, list) or isinstance(command, tuple):
            command = " ".join([six.text_type(c) for c in command])

        sanitized_cmd = strutils.mask_password(command)

        b64_command = base64.b64encode(command.encode("utf_16_le"))
        command = ("powershell.exe -ExecutionPolicy RemoteSigned "
                   "-NonInteractive -EncodedCommand %s" % b64_command)
        return command, sanitized_cmd

    def _get_auth(self, server):
        auth = {'username': server['username']}

        if server['use_cert_auth']:
            auth['cert_pem_path'] = server['cert_pem_path']
            auth['cert_key_pem_path'] = server['cert_key_pem_path']
        else:
            auth['password'] = server['password']
        return auth


class WinRMConnection(object):
    _URL_TEMPLATE = '%(protocol)s://%(ip)s:%(port)s/wsman'

    def __init__(self, ip=None, port=None, use_ssl=False,
                 transport=None, username=None, password=None,
                 cert_pem_path=None, cert_key_pem_path=None,
                 operation_timeout=None, conn_timeout=None):
        setup_winrm()

        use_cert = bool(cert_pem_path and cert_key_pem_path)
        transport = (TRANSPORT_SSL
                     if use_cert else TRANSPORT_PLAINTEXT)

        _port = port or self._get_default_port(use_cert)
        _url = self._get_url(ip, _port, use_cert)

        self._conn = winrm.protocol.Protocol(
            endpoint=_url, transport=transport,
            username=username, password=password,
            cert_pem=cert_pem_path, cert_key_pem=cert_key_pem_path)
        self._conn.transport.timeout = conn_timeout
        self._conn.set_timeout(operation_timeout)

    def _get_default_port(self, use_ssl):
        port = (DEFAULT_PORT_HTTPS
                if use_ssl else DEFAULT_PORT_HTTP)
        return port

    def _get_url(self, ip, port, use_ssl):
        if not ip:
            err_msg = _("No IP provided.")
            raise exception.ShareBackendException(msg=err_msg)

        protocol = 'https' if use_ssl else 'http'
        return self._URL_TEMPLATE % {'protocol': protocol,
                                     'ip': ip,
                                     'port': port}

    def execute(self, cmd):
        shell_id = None
        cmd_id = None

        try:
            shell_id = self._conn.open_shell()

            cmd_id = self._conn.run_command(shell_id, cmd)

            (stdout,
             stderr,
             exit_code) = self._conn.get_command_output(shell_id, cmd_id)
        finally:
            if cmd_id:
                self._conn.cleanup_command(shell_id, cmd_id)
            if shell_id:
                self._conn.close_shell(shell_id)

        return (stdout, stderr, exit_code)



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ssl
from unittest import mock

import ddt

from manila.share.drivers.dell_emc.common.enas import utils
from manila import test


@ddt.ddt
class ENASUtilsTestCase(test.TestCase):

    @ddt.data({'full': ['cge-1-0', 'cge-1-1', 'cge-3-0',
                        'cge-3-1', 'cge-12-3'],
               'matchers': ['cge-?-0', 'cge-3*', 'foo'],
               'matched': set(['cge-1-0', 'cge-3-0',
                               'cge-3-1']),
               'unmatched': set(['cge-1-1', 'cge-12-3'])},
              {'full': ['cge-1-0', 'cge-1-1'],
               'matchers': ['cge-1-0'],
               'matched': set(['cge-1-0']),
               'unmatched': set(['cge-1-1'])},
              {'full': ['cge-1-0', 'cge-1-1'],
               'matchers': ['foo'],
               'matched': set([]),
               'unmatched': set(['cge-1-0', 'cge-1-1'])})
    @ddt.unpack
    def test_do_match_any(self, full, matchers, matched, unmatched):
        real_matched, real_unmatched = utils.do_match_any(
            full, matchers)
        self.assertEqual(matched, real_matched)
        self.assertEqual(unmatched, real_unmatched)


class SslContextTestCase(test.TestCase):

    def test_create_ssl_context(self):
        configuration = mock.Mock()
        configuration.emc_ssl_cert_verify = True
        configuration.emc_ssl_cert_path = "./cert_path/"
        self.mock_object(ssl, 'create_default_context')
        context = utils.create_ssl_context(configuration)
        self.assertIsNotNone(context)

    def test_create_ssl_context_no_verify(self):
        configuration = mock.Mock()
        configuration.emc_ssl_cert_verify = False
        self.mock_object(ssl, 'create_default_context')
        context = utils.create_ssl_context(configuration)
        self.assertFalse(context.check_hostname)

    def test_no_create_default_context(self):
        """Test scenario of running on python 2.7.8 or earlier."""
        configuration = mock.Mock()
        configuration.emc_ssl_cert_verify = False
        self.mock_object(ssl, 'create_default_context',
                         mock.Mock(side_effect=AttributeError))
        context = utils.create_ssl_context(configuration)
        self.assertIsNone(context)


@ddt.ddt
class ParseIpaddrTestCase(test.TestCase):

    @ddt.data({'lst_ipaddr': ['192.168.100.101',
                              '192.168.100.102',
                              '192.168.100.103']},
              {'lst_ipaddr': ['[fdf8:f53b:82e4::57]',
                              '[fdf8:f53b:82e4::54]',
                              '[fdf8:f53b:82e4::55]']},
              {'lst_ipaddr': ['[fdf8:f53b:82e4::57]',
                              '[fdf8:f53b:82e4::54]',
                              '192.168.100.103',
                              '[fdf8:f53b:82e4::55]']},
              {'lst_ipaddr': ['192.168.100.101',
                              '[fdf8:f53b:82e4::57]',
                              '[fdf8:f53b:82e4::54]',
                              '192.168.100.101',
                              '[fdf8:f53b:82e4::55]',
                              '192.168.100.102']},)
    @ddt.unpack
    def test_parse_ipv4_addr(self, lst_ipaddr):
        self.assertEqual(lst_ipaddr, utils.parse_ipaddr(':'.join(lst_ipaddr)))


@ddt.ddt
class ConvertIPv6FormatTestCase(test.TestCase):

    @ddt.data({'ip_addr': 'fdf8:f53b:82e4::55'},
              {'ip_addr': 'fdf8:f53b:82e4::55/64'},
              {'ip_addr': 'fdf8:f53b:82e4::55/128'})
    @ddt.unpack
    def test_ipv6_addr(self, ip_addr):
        expected_ip_addr = '[%s]' % ip_addr
        self.assertEqual(expected_ip_addr,
                         utils.convert_ipv6_format_if_needed(ip_addr))

    @ddt.data({'ip_addr': '192.168.1.100'},
              {'ip_addr': '192.168.1.100/24'},
              {'ip_addr': '192.168.1.100/32'},
              {'ip_addr': '[fdf8:f53b:82e4::55]'})
    @ddt.unpack
    def test_invalid_ipv6_addr(self, ip_addr):
        self.assertEqual(ip_addr, utils.convert_ipv6_format_if_needed(ip_addr))


@ddt.ddt
class ExportUncPathTestCase(test.TestCase):

    @ddt.data({'ip_addr': 'fdf8:f53b:82e4::55'},
              {'ip_addr': 'fdf8:f53b:82e4::'},
              {'ip_addr': '2018::'})
    @ddt.unpack
    def test_ipv6_addr(self, ip_addr):
        expected_ip_addr = '%s.ipv6-literal.net' % ip_addr.replace(':', '-')
        self.assertEqual(expected_ip_addr,
                         utils.export_unc_path(ip_addr))

    @ddt.data({'ip_addr': '192.168.1.100'},
              {'ip_addr': '192.168.1.100/24'},
              {'ip_addr': '192.168.1.100/32'},
              {'ip_addr': 'fdf8:f53b:82e4::55/64'},
              {'ip_addr': 'fdf8:f53b:82e4::55/128'},
              {'ip_addr': '[fdf8:f53b:82e4::55]'})
    @ddt.unpack
    def test_invalid_ipv6_addr(self, ip_addr):
        self.assertEqual(ip_addr, utils.export_unc_path(ip_addr))


@ddt.ddt
class SizeToGbTestCase(test.TestCase):

    @ddt.data({'size_in_bytes': 1073741824, 'size_in_gb': 1.0},
              {'size_in_bytes': 5610301030, 'size_in_gb': 5.22})
    @ddt.unpack
    def test_bytes_to_gb(self, size_in_bytes, size_in_gb):
        self.assertEqual(size_in_gb, utils.bytes_to_gb(size_in_bytes))

    @ddt.data({'size_in_mb': 1024, 'size_in_gb': 1.0},
              {'size_in_mb': 5346, 'size_in_gb': 5.22})
    @ddt.unpack
    def test_mb_to_gb(self, size_in_mb, size_in_gb):
        self.assertEqual(size_in_gb, utils.mb_to_gb(size_in_mb))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

from oslo_config import cfg
from oslo_utils import timeutils

from manila import context
from manila.message import api as message_api
from manila.message.message_field import Action as MsgAction
from manila.message.message_field import Detail as MsgDetail
from manila.message import message_levels
from manila import test

CONF = cfg.CONF


class MessageApiTest(test.TestCase):
    def setUp(self):
        super(MessageApiTest, self).setUp()
        self.message_api = message_api.API()
        self.mock_object(self.message_api, 'db')
        self.ctxt = context.RequestContext('admin', 'fakeproject', True)
        self.ctxt.request_id = 'fakerequestid'

    @mock.patch.object(timeutils, 'utcnow')
    def test_create(self, mock_utcnow):
        CONF.set_override('message_ttl', 300)
        now = datetime.datetime.utcnow()
        mock_utcnow.return_value = now
        expected_expires_at = now + datetime.timedelta(
            seconds=300)
        expected_message_record = {
            'project_id': 'fakeproject',
            'request_id': 'fakerequestid',
            'resource_type': 'fake_resource_type',
            'resource_id': None,
            'action_id': MsgAction.ALLOCATE_HOST[0],
            'detail_id': MsgDetail.NO_VALID_HOST[0],
            'message_level': message_levels.ERROR,
            'expires_at': expected_expires_at,
        }

        self.message_api.create(self.ctxt,
                                MsgAction.ALLOCATE_HOST,
                                "fakeproject",
                                detail=MsgDetail.NO_VALID_HOST,
                                resource_type="fake_resource_type")

        self.message_api.db.message_create.assert_called_once_with(
            self.ctxt, expected_message_record)

    def test_create_swallows_exception(self):
        self.mock_object(self.message_api.db, 'message_create',
                         mock.Mock(side_effect=Exception()))
        exception_log = self.mock_object(message_api.LOG, 'exception')
        self.message_api.create(self.ctxt,
                                MsgAction.ALLOCATE_HOST,
                                'fakeproject',
                                'fake_resource')

        self.message_api.db.message_create.assert_called_once_with(
            self.ctxt, mock.ANY)
        exception_log.assert_called_once_with(
            'Failed to create message record for request_id %s',
            self.ctxt.request_id)

    def test_get(self):
        self.message_api.get(self.ctxt, 'fake_id')

        self.message_api.db.message_get.assert_called_once_with(self.ctxt,
                                                                'fake_id')

    def test_get_all(self):
        self.message_api.get_all(self.ctxt)

        self.message_api.db.message_get_all.assert_called_once_with(
            self.ctxt, filters={}, limit=None, offset=None,
            sort_dir=None, sort_key=None)

    def test_delete(self):
        self.message_api.delete(self.ctxt, 'fake_id')

        self.message_api.db.message_destroy.assert_called_once_with(
            self.ctxt, 'fake_id')

    def test_cleanup_expired_messages(self):
        admin_context = mock.Mock()
        self.mock_object(self.ctxt, 'elevated',
                         mock.Mock(return_value=admin_context))
        self.message_api.cleanup_expired_messages(self.ctxt)
        self.message_api.db.cleanup_expired_messages.assert_called_once_with(
            admin_context)



#!/usr/bin/env python

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Starter script for manila OS API."""

import eventlet
eventlet.monkey_patch()
# Monkey patch the original current_thread to use the up-to-date _active
# global variable. See https://bugs.launchpad.net/bugs/1863021 and
# https://github.com/eventlet/eventlet/issues/592
import __original_module_threading as orig_threading
import threading  # noqa
orig_threading.current_thread.__globals__['_active'] = threading._active

import sys

from oslo_config import cfg
from oslo_log import log
from oslo_reports import guru_meditation_report as gmr
from oslo_reports import opts as gmr_opts

from manila.common import config  # Need to register global_opts  # noqa
from manila import service
from manila import utils
from manila import version

CONF = cfg.CONF


def main():
    log.register_options(CONF)
    gmr_opts.set_defaults(CONF)
    CONF(sys.argv[1:], project='manila',
         version=version.version_string())
    config.verify_share_protocols()
    log.setup(CONF, "manila")
    utils.monkey_patch()

    gmr.TextGuruMeditation.setup_autorun(version, conf=CONF)
    launcher = service.process_launcher()
    server = service.WSGIService('osapi_share')
    launcher.launch_service(server, workers=server.workers or 1)
    launcher.wait()


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_share_instances_access_map

Revision ID: 579c267fbb4d
Revises: 5077ffcc5f1c
Create Date: 2015-08-19 07:51:52.928542

"""

# revision identifiers, used by Alembic.
revision = '579c267fbb4d'
down_revision = '5077ffcc5f1c'

from alembic import op
from sqlalchemy import Column, DateTime, ForeignKey, String
from oslo_utils import uuidutils

from manila.db.migrations import utils


def upgrade():
    """Create 'share_instance_access_map' table and move 'state' column."""

    instance_access_table = op.create_table(
        'share_instance_access_map',
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('deleted_at', DateTime),
        Column('deleted', String(length=36), default='False'),
        Column('id', String(length=36), primary_key=True, nullable=False),
        Column('share_instance_id', String(length=36),
               ForeignKey('share_instances.id', name="siam_instance_fk")),
        Column('access_id', String(length=36),
               ForeignKey('share_access_map.id', name="siam_access_fk")),
        Column('state', String(length=255)),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    # NOTE(u_glide): Move all states from 'share_access_map'
    # to 'share_instance_access_map'
    instance_access_mappings = []
    connection = op.get_bind()
    access_table = utils.load_table('share_access_map', connection)
    instances_table = utils.load_table('share_instances', connection)

    for access_rule in connection.execute(access_table.select()):
        # pylint: disable=assignment-from-no-return
        instances_query = instances_table.select().where(
            instances_table.c.share_id == access_rule.share_id
        )

        for instance in connection.execute(instances_query):
            instance_access_mappings.append({
                'created_at': access_rule.created_at,
                'updated_at': access_rule.updated_at,
                'deleted_at': access_rule.deleted_at,
                'deleted': access_rule.deleted,
                'id': uuidutils.generate_uuid(),
                'share_instance_id': instance.id,
                'access_id': access_rule.id,
                'state': access_rule.state,
            })
    op.bulk_insert(instance_access_table, instance_access_mappings)
    op.drop_column('share_access_map', 'state')


def downgrade():
    """Remove 'share_instance_access_map' table and add 'state' column back.

    This method can lead to data loss because only first state is saved in
    share_access_map table.
    """
    op.add_column('share_access_map', Column('state', String(length=255)))

    # NOTE(u_glide): Move all states from 'share_instance_access_map'
    # to 'share_access_map'
    connection = op.get_bind()
    access_table = utils.load_table('share_access_map', connection)
    instance_access_table = utils.load_table('share_instance_access_map',
                                             connection)

    share_access_rules = connection.execute(
        access_table.select().where(access_table.c.deleted == "False"))

    for access_rule in share_access_rules:
        access_mapping = connection.execute(
            instance_access_table.select().where(
                instance_access_table.c.access_id == access_rule['id'])
        ).first()

        # pylint: disable=no-value-for-parameter
        op.execute(
            access_table.update().where(
                access_table.c.id == access_rule['id']
            ).values({'state': access_mapping['state']})
        )

    op.drop_table('share_instance_access_map')



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for the ShareReplicationFilter.
"""
import ddt

from oslo_context import context

from manila.scheduler.filters import share_replication
from manila import test
from manila.tests.scheduler import fakes


@ddt.ddt
class ShareReplicationFilterTestCase(test.TestCase):
    """Test case for ShareReplicationFilter."""

    def setUp(self):
        super(ShareReplicationFilterTestCase, self).setUp()
        self.filter = share_replication.ShareReplicationFilter()
        self.debug_log = self.mock_object(share_replication.LOG, 'debug')

    @staticmethod
    def _create_replica_request(replication_domain='kashyyyk',
                                replication_type='dr',
                                active_replica_host=fakes.FAKE_HOST_STRING_1,
                                all_replica_hosts=fakes.FAKE_HOST_STRING_1,
                                is_admin=False):
        ctxt = context.RequestContext('fake', 'fake', is_admin=is_admin)
        return {
            'context': ctxt,
            'request_spec': {
                'active_replica_host': active_replica_host,
                'all_replica_hosts': all_replica_hosts,
            },
            'resource_type': {
                'extra_specs': {
                    'replication_type': replication_type,
                },
            },
            'replication_domain': replication_domain,
        }

    @ddt.data('tatooine', '')
    def test_share_replication_filter_fails_incompatible_domain(self, domain):
        request = self._create_replica_request()

        host = fakes.FakeHostState('host1',
                                   {
                                       'replication_domain': domain,
                                   })

        self.assertFalse(self.filter.host_passes(host, request))
        self.assertTrue(self.debug_log.called)

    def test_share_replication_filter_fails_no_replication_domain(self):
        request = self._create_replica_request()

        host = fakes.FakeHostState('host1',
                                   {
                                       'replication_domain': None,
                                   })

        self.assertFalse(self.filter.host_passes(host, request))
        self.assertTrue(self.debug_log.called)

    def test_share_replication_filter_fails_host_has_replicas(self):
        all_replica_hosts = ','.join(['host1', fakes.FAKE_HOST_STRING_1])
        request = self._create_replica_request(
            all_replica_hosts=all_replica_hosts)

        host = fakes.FakeHostState('host1',
                                   {
                                       'replication_domain': 'kashyyyk',
                                   })
        self.assertFalse(self.filter.host_passes(host, request))
        self.assertTrue(self.debug_log.called)

    def test_share_replication_filter_passes_no_replication_type(self):
        request = self._create_replica_request(replication_type=None)

        host = fakes.FakeHostState('host1',
                                   {
                                       'replication_domain': 'tatooine',
                                   })

        self.assertTrue(self.filter.host_passes(host, request))

    def test_share_replication_filter_passes_no_active_replica_host(self):
        request = self._create_replica_request(active_replica_host=None)

        host = fakes.FakeHostState('host1',
                                   {
                                       'replication_domain': 'tatooine',
                                   })

        self.assertTrue(self.filter.host_passes(host, request))

    def test_share_replication_filter_passes_happy_day(self):
        all_replica_hosts = ','.join(['host1', fakes.FAKE_HOST_STRING_1])
        request = self._create_replica_request(
            all_replica_hosts=all_replica_hosts)

        host = fakes.FakeHostState('host2',
                                   {
                                       'replication_domain': 'kashyyyk',
                                   })

        self.assertTrue(self.filter.host_passes(host, request))

    def test_share_replication_filter_empty(self):
        request = {}

        host = fakes.FakeHostState('host1',
                                   {
                                       'replication_domain': 'naboo',
                                   })

        self.assertTrue(self.filter.host_passes(host, request))



#    Copyright 2014 Mirantis Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from manila.tests import fake_compute


class FakeServiceInstanceManager(object):

    def __init__(self, *args, **kwargs):
        self.db = mock.Mock()
        self._helpers = {
            'CIFS': mock.Mock(),
            'NFS': mock.Mock(),
        }
        self.share_networks_locks = {}
        self.share_networks_servers = {}
        self.fake_server = fake_compute.FakeServer()
        self.service_instance_name_template = 'manila_fake_service_instance-%s'
        self._network_helper = None

    def get_service_instance(self, context, share_network_id, create=True):
        return self.fake_server

    @property
    def network_helper(self):
        return self._get_network_helper()

    def _get_network_helper(self):
        self._network_helper = FakeNeutronNetworkHelper()
        return self._network_helper

    def _create_service_instance(self, context, instance_name,
                                 share_network_id, old_server_ip):
        return self.fake_server

    def _delete_server(self, context, server):
        pass

    def _get_service_instance_name(self, share_network_id):
        return self.service_instance_name_template % share_network_id


class FakeNeutronNetworkHelper(object):

    def setup_connectivity_with_service_instances(self):
        pass



# Copyright (c) 2011 OpenStack, LLC
# Copyright (c) 2015 Rushil Chugh
# Copyright (c) 2015 Clinton Knight
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For HostManager
"""

import copy
from unittest import mock

import ddt
from oslo_config import cfg
from oslo_utils import timeutils

from manila import context
from manila import db
from manila import exception
from manila.scheduler.filters import base_host
from manila.scheduler import host_manager
from manila.scheduler import utils as scheduler_utils
from manila import test
from manila.tests.scheduler import fakes
from manila import utils


CONF = cfg.CONF


class FakeFilterClass1(base_host.BaseHostFilter):
    def host_passes(self, host_state, filter_properties):
        pass


class FakeFilterClass2(base_host.BaseHostFilter):
    def host_passes(self, host_state, filter_properties):
        pass


@ddt.ddt
class HostManagerTestCase(test.TestCase):
    """Test case for HostManager class."""

    def setUp(self):
        super(HostManagerTestCase, self).setUp()
        self.host_manager = host_manager.HostManager()
        self.fake_hosts = [host_manager.HostState('fake_host%s' % x)
                           for x in range(1, 5)]

    def test_choose_host_filters_not_found(self):
        self.flags(scheduler_default_filters='FakeFilterClass3')
        self.host_manager.filter_classes = [FakeFilterClass1,
                                            FakeFilterClass2]
        self.assertRaises(exception.SchedulerHostFilterNotFound,
                          self.host_manager._choose_host_filters, None)

    def test_choose_host_filters(self):
        self.flags(scheduler_default_filters=['FakeFilterClass2'])
        self.host_manager.filter_classes = [FakeFilterClass1,
                                            FakeFilterClass2]

        # Test 'share' returns 1 correct function
        filter_classes = self.host_manager._choose_host_filters(None)
        self.assertEqual(1, len(filter_classes))
        self.assertEqual('FakeFilterClass2', filter_classes[0].__name__)

    def _verify_result(self, info, result):
        for x in info['got_fprops']:
            self.assertEqual(info['expected_fprops'], x)
        self.assertEqual(set(info['expected_objs']), set(info['got_objs']))
        self.assertEqual(set(info['got_objs']), set(result))

    def test_get_filtered_hosts(self):
        fake_properties = {'moo': 1, 'cow': 2}
        info = {
            'expected_objs': self.fake_hosts,
            'expected_fprops': fake_properties,
        }
        with mock.patch.object(self.host_manager, '_choose_host_filters',
                               mock.Mock(return_value=[FakeFilterClass1])):
            info['got_objs'] = []
            info['got_fprops'] = []

            def fake_filter_one(_self, obj, filter_props):
                info['got_objs'].append(obj)
                info['got_fprops'].append(filter_props)
                return True

            self.mock_object(FakeFilterClass1, '_filter_one', fake_filter_one)
            result, last_filter = self.host_manager.get_filtered_hosts(
                self.fake_hosts, fake_properties)
            self._verify_result(info, result)
            self.host_manager._choose_host_filters.assert_called_once_with(
                mock.ANY)

    def test_update_service_capabilities_for_shares(self):
        service_states = self.host_manager.service_states
        self.assertDictEqual(service_states, {})
        host1_share_capabs = dict(free_capacity_gb=4321, timestamp=1)
        host2_share_capabs = dict(free_capacity_gb=5432, timestamp=1)
        host3_share_capabs = dict(free_capacity_gb=6543, timestamp=1)
        service_name = 'share'
        self.host_manager.update_service_capabilities(
            service_name, 'host1', host1_share_capabs, 31337)

        self.host_manager.update_service_capabilities(
            service_name, 'host2', host2_share_capabs, 31338)

        self.host_manager.update_service_capabilities(
            service_name, 'host3', host3_share_capabs, 31339)

        # Make sure dictionary isn't re-assigned
        self.assertEqual(service_states, self.host_manager.service_states)
        # Make sure original dictionary wasn't copied
        self.assertEqual(1, host1_share_capabs['timestamp'])

        host1_share_capabs['timestamp'] = 31337
        host2_share_capabs['timestamp'] = 31338
        host3_share_capabs['timestamp'] = 31339

        expected = {
            'host1': host1_share_capabs,
            'host2': host2_share_capabs,
            'host3': host3_share_capabs,
        }
        self.assertDictEqual(service_states, expected)

    def test_get_all_host_states_share(self):
        fake_context = context.RequestContext('user', 'project')
        topic = CONF.share_topic
        tmp_pools = copy.deepcopy(fakes.SHARE_SERVICES_WITH_POOLS)
        tmp_enable_pools = tmp_pools[:-2]
        self.mock_object(
            db, 'service_get_all_by_topic',
            mock.Mock(return_value=tmp_enable_pools))
        self.mock_object(utils, 'service_is_up', mock.Mock(return_value=True))

        with mock.patch.dict(self.host_manager.service_states,
                             fakes.SHARE_SERVICE_STATES_WITH_POOLS):
            # Get service
            self.host_manager.get_all_host_states_share(fake_context)

            # Disabled one service
            tmp_enable_pools.pop()
            self.mock_object(
                db, 'service_get_all_by_topic',
                mock.Mock(return_value=tmp_enable_pools))

            # Get service again
            self.host_manager.get_all_host_states_share(fake_context)
            host_state_map = self.host_manager.host_state_map

            self.assertEqual(3, len(host_state_map))
            # Check that service is up
            for i in range(3):
                share_node = fakes.SHARE_SERVICES_WITH_POOLS[i]
                host = share_node['host']
                self.assertEqual(share_node, host_state_map[host].service)
            db.service_get_all_by_topic.assert_called_once_with(
                fake_context, topic)

    def test_get_pools_no_pools(self):
        fake_context = context.RequestContext('user', 'project')
        self.mock_object(utils, 'service_is_up', mock.Mock(return_value=True))
        self.mock_object(
            db, 'service_get_all_by_topic',
            mock.Mock(return_value=fakes.SHARE_SERVICES_NO_POOLS))
        host_manager.LOG.warning = mock.Mock()

        with mock.patch.dict(self.host_manager.service_states,
                             fakes.SERVICE_STATES_NO_POOLS):

            res = self.host_manager.get_pools(context=fake_context)

            expected = [
                {
                    'name': 'host1#AAA',
                    'host': 'host1',
                    'backend': None,
                    'pool': 'AAA',
                    'capabilities': {
                        'timestamp': None,
                        'share_backend_name': 'AAA',
                        'free_capacity_gb': 200,
                        'driver_version': None,
                        'total_capacity_gb': 512,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 312,
                        'max_over_subscription_ratio': 1.0,
                        'thin_provisioning': False,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': False,
                        'create_share_from_snapshot_support': False,
                        'revert_to_snapshot_support': True,
                        'mount_snapshot_support': True,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                }, {
                    'name': 'host2@back1#BBB',
                    'host': 'host2',
                    'backend': 'back1',
                    'pool': 'BBB',
                    'capabilities': {
                        'timestamp': None,
                        'share_backend_name': 'BBB',
                        'free_capacity_gb': 100,
                        'driver_version': None,
                        'total_capacity_gb': 256,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 400,
                        'max_over_subscription_ratio': 2.0,
                        'thin_provisioning': True,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                }, {
                    'name': 'host2@back2#CCC',
                    'host': 'host2',
                    'backend': 'back2',
                    'pool': 'CCC',
                    'capabilities': {
                        'timestamp': None,
                        'share_backend_name': 'CCC',
                        'free_capacity_gb': 700,
                        'driver_version': None,
                        'total_capacity_gb': 10000,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 50000,
                        'max_over_subscription_ratio': 20.0,
                        'thin_provisioning': True,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                },
            ]
            self.assertIsInstance(res, list)
            self.assertEqual(len(expected), len(res))
            for pool in expected:
                self.assertIn(pool, res)

    def test_get_pools(self):
        fake_context = context.RequestContext('user', 'project')
        self.mock_object(utils, 'service_is_up', mock.Mock(return_value=True))
        self.mock_object(
            db, 'service_get_all_by_topic',
            mock.Mock(return_value=fakes.SHARE_SERVICES_WITH_POOLS))
        host_manager.LOG.warning = mock.Mock()

        with mock.patch.dict(self.host_manager.service_states,
                             fakes.SHARE_SERVICE_STATES_WITH_POOLS):

            res = self.host_manager.get_pools(fake_context)

            expected = [
                {
                    'name': 'host1@AAA#pool1',
                    'host': 'host1',
                    'backend': 'AAA',
                    'pool': 'pool1',
                    'capabilities': {
                        'pool_name': 'pool1',
                        'timestamp': None,
                        'share_backend_name': 'AAA',
                        'free_capacity_gb': 41,
                        'driver_version': None,
                        'total_capacity_gb': 51,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 10,
                        'max_over_subscription_ratio': 1.0,
                        'thin_provisioning': False,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': True,
                        'mount_snapshot_support': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                }, {
                    'name': 'host2@BBB#pool2',
                    'host': 'host2',
                    'backend': 'BBB',
                    'pool': 'pool2',
                    'capabilities': {
                        'pool_name': 'pool2',
                        'timestamp': None,
                        'share_backend_name': 'BBB',
                        'free_capacity_gb': 42,
                        'driver_version': None,
                        'total_capacity_gb': 52,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 60,
                        'max_over_subscription_ratio': 2.0,
                        'thin_provisioning': True,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                }, {
                    'name': 'host3@CCC#pool3',
                    'host': 'host3',
                    'backend': 'CCC',
                    'pool': 'pool3',
                    'capabilities': {
                        'pool_name': 'pool3',
                        'timestamp': None,
                        'share_backend_name': 'CCC',
                        'free_capacity_gb': 43,
                        'driver_version': None,
                        'total_capacity_gb': 53,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 100,
                        'max_over_subscription_ratio': 20.0,
                        'thin_provisioning': True,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                }, {
                    'name': 'host4@DDD#pool4a',
                    'host': 'host4',
                    'backend': 'DDD',
                    'pool': 'pool4a',
                    'capabilities': {
                        'pool_name': 'pool4a',
                        'timestamp': None,
                        'share_backend_name': 'DDD',
                        'free_capacity_gb': 441,
                        'driver_version': None,
                        'total_capacity_gb': 541,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 800,
                        'max_over_subscription_ratio': 2.0,
                        'thin_provisioning': True,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                }, {
                    'name': 'host4@DDD#pool4b',
                    'host': 'host4',
                    'backend': 'DDD',
                    'pool': 'pool4b',
                    'capabilities': {
                        'pool_name': 'pool4b',
                        'timestamp': None,
                        'share_backend_name': 'DDD',
                        'free_capacity_gb': 442,
                        'driver_version': None,
                        'total_capacity_gb': 542,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 2000,
                        'max_over_subscription_ratio': 10.0,
                        'thin_provisioning': True,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                },
            ]
            self.assertIsInstance(res, list)
            self.assertIsInstance(self.host_manager.host_state_map, dict)
            self.assertEqual(len(expected), len(res))
            for pool in expected:
                self.assertIn(pool, res)

    def test_get_pools_host_down(self):
        fake_context = context.RequestContext('user', 'project')
        mock_service_is_up = self.mock_object(utils, 'service_is_up')
        self.mock_object(
            db, 'service_get_all_by_topic',
            mock.Mock(return_value=fakes.SHARE_SERVICES_NO_POOLS))
        host_manager.LOG.warning = mock.Mock()

        with mock.patch.dict(self.host_manager.service_states,
                             fakes.SERVICE_STATES_NO_POOLS):

            # Initialize host data with all services present
            mock_service_is_up.side_effect = [True, True, True]

            # Call once to update the host state map
            self.host_manager.get_pools(fake_context)

            self.assertEqual(len(fakes.SHARE_SERVICES_NO_POOLS),
                             len(self.host_manager.host_state_map))

            # Then mock one host as down
            mock_service_is_up.side_effect = [True, True, False]

            res = self.host_manager.get_pools(fake_context)

            expected = [
                {
                    'name': 'host1#AAA',
                    'host': 'host1',
                    'backend': None,
                    'pool': 'AAA',
                    'capabilities': {
                        'timestamp': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': False,
                        'create_share_from_snapshot_support': False,
                        'revert_to_snapshot_support': True,
                        'mount_snapshot_support': True,
                        'share_backend_name': 'AAA',
                        'free_capacity_gb': 200,
                        'driver_version': None,
                        'total_capacity_gb': 512,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'provisioned_capacity_gb': 312,
                        'max_over_subscription_ratio': 1.0,
                        'thin_provisioning': False,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                }, {
                    'name': 'host2@back1#BBB',
                    'host': 'host2',
                    'backend': 'back1',
                    'pool': 'BBB',
                    'capabilities': {
                        'timestamp': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'share_backend_name': 'BBB',
                        'free_capacity_gb': 100,
                        'driver_version': None,
                        'total_capacity_gb': 256,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'provisioned_capacity_gb': 400,
                        'max_over_subscription_ratio': 2.0,
                        'thin_provisioning': True,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                },
            ]
            self.assertIsInstance(res, list)
            self.assertIsInstance(self.host_manager.host_state_map, dict)
            self.assertEqual(len(expected), len(res))
            self.assertEqual(len(expected),
                             len(self.host_manager.host_state_map))
            for pool in expected:
                self.assertIn(pool, res)

    def test_get_pools_with_filters(self):
        fake_context = context.RequestContext('user', 'project')
        self.mock_object(utils, 'service_is_up', mock.Mock(return_value=True))
        self.mock_object(
            db, 'service_get_all_by_topic',
            mock.Mock(return_value=fakes.SHARE_SERVICES_WITH_POOLS))
        host_manager.LOG.warning = mock.Mock()

        with mock.patch.dict(self.host_manager.service_states,
                             fakes.SHARE_SERVICE_STATES_WITH_POOLS):

            res = self.host_manager.get_pools(
                context=fake_context,
                filters={'host': 'host2', 'pool': 'pool*',
                         'capabilities': {'dedupe': 'False'}})

            expected = [
                {
                    'name': 'host2@BBB#pool2',
                    'host': 'host2',
                    'backend': 'BBB',
                    'pool': 'pool2',
                    'capabilities': {
                        'pool_name': 'pool2',
                        'timestamp': None,
                        'driver_handles_share_servers': False,
                        'snapshot_support': True,
                        'create_share_from_snapshot_support': True,
                        'revert_to_snapshot_support': False,
                        'mount_snapshot_support': False,
                        'share_backend_name': 'BBB',
                        'free_capacity_gb': 42,
                        'driver_version': None,
                        'total_capacity_gb': 52,
                        'reserved_percentage': 0,
                        'reserved_snapshot_percentage': 0,
                        'provisioned_capacity_gb': 60,
                        'max_over_subscription_ratio': 2.0,
                        'thin_provisioning': True,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'dedupe': False,
                        'compression': False,
                        'replication_type': None,
                        'replication_domain': None,
                        'sg_consistent_snapshot_support': None,
                        'security_service_update_support': False,
                    },
                },
            ]
            self.assertTrue(isinstance(res, list))
            self.assertEqual(len(expected), len(res))
            for pool in expected:
                self.assertIn(pool, res)

    @ddt.data(
        None,
        {},
        {'key1': 'value1'},
        {'capabilities': {'dedupe': 'False'}},
        {'capabilities': {'dedupe': '<is> False'}},
        {'key1': 'value1', 'key2': 'value*'},
        {'key1': '.*', 'key2': '.*'},
    )
    def test_passes_filters_true(self, filter):

        data = {
            'key1': 'value1',
            'key2': 'value2',
            'key3': 'value3',
            'capabilities': {'dedupe': False},
        }
        self.assertTrue(self.host_manager._passes_filters(data, filter))

    @ddt.data(
        {'key1': 'value$'},
        {'key4': 'value'},
        {'capabilities': {'dedupe': 'True'}},
        {'capabilities': {'dedupe': '<is> True'}},
        {'key1': 'value1.+', 'key2': 'value*'},
    )
    def test_passes_filters_false(self, filter):

        data = {
            'key1': 'value1',
            'key2': 'value2',
            'key3': 'value3',
            'capabilities': {'dedupe': False},
        }
        self.assertFalse(self.host_manager._passes_filters(data, filter))


class HostStateTestCase(test.TestCase):
    """Test case for HostState class."""

    def test_update_from_share_capability_nopool(self):
        fake_context = context.RequestContext('user', 'project', is_admin=True)
        share_capability = {'total_capacity_gb': 0,
                            'free_capacity_gb': 100,
                            'reserved_percentage': 0,
                            'reserved_snapshot_percentage': 0,
                            'timestamp': None,
                            'ipv4_support': True,
                            'ipv6_support': False}
        fake_host = host_manager.HostState('host1', share_capability)
        self.assertIsNone(fake_host.free_capacity_gb)

        fake_host.update_from_share_capability(share_capability,
                                               context=fake_context)
        # Backend level stats remain uninitialized
        self.assertEqual(0, fake_host.total_capacity_gb)
        self.assertIsNone(fake_host.free_capacity_gb)
        self.assertTrue(fake_host.ipv4_support)
        self.assertFalse(fake_host.ipv6_support)
        # Pool stats has been updated
        self.assertEqual(0, fake_host.pools['_pool0'].total_capacity_gb)
        self.assertEqual(100, fake_host.pools['_pool0'].free_capacity_gb)
        self.assertTrue(fake_host.pools['_pool0'].ipv4_support)
        self.assertFalse(fake_host.pools['_pool0'].ipv6_support)

        # Test update for existing host state
        share_capability.update(dict(total_capacity_gb=1000))
        fake_host.update_from_share_capability(share_capability,
                                               context=fake_context)
        self.assertEqual(1000, fake_host.pools['_pool0'].total_capacity_gb)

        # Test update for existing host state with different backend name
        share_capability.update(dict(share_backend_name='magic'))
        fake_host.update_from_share_capability(share_capability,
                                               context=fake_context)
        self.assertEqual(1000, fake_host.pools['magic'].total_capacity_gb)
        self.assertEqual(100, fake_host.pools['magic'].free_capacity_gb)
        # 'pool0' becomes nonactive pool, and is deleted
        self.assertRaises(KeyError, lambda: fake_host.pools['pool0'])

    def test_update_from_share_capability_with_pools(self):
        fake_context = context.RequestContext('user', 'project', is_admin=True)
        fake_host = host_manager.HostState('host1#pool1')
        self.assertIsNone(fake_host.free_capacity_gb)
        capability = {
            'share_backend_name': 'Backend1',
            'vendor_name': 'OpenStack',
            'driver_version': '1.1',
            'storage_protocol': 'NFS_CIFS',
            'ipv4_support': True,
            'ipv6_support': False,
            'pools': [
                {'pool_name': 'pool1',
                 'total_capacity_gb': 500,
                 'free_capacity_gb': 230,
                 'allocated_capacity_gb': 270,
                 'qos': 'False',
                 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'dying_disks': 100,
                 'super_hero_1': 'spider-man',
                 'super_hero_2': 'flash',
                 'super_hero_3': 'neoncat',
                 },
                {'pool_name': 'pool2',
                 'total_capacity_gb': 1024,
                 'free_capacity_gb': 1024,
                 'allocated_capacity_gb': 0,
                 'qos': 'False',
                 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'dying_disks': 200,
                 'super_hero_1': 'superman',
                 'super_hero_2': 'Hulk',
                 }
            ],
            'timestamp': None,
        }

        fake_host.update_from_share_capability(capability,
                                               context=fake_context)

        self.assertEqual('Backend1', fake_host.share_backend_name)
        self.assertEqual('NFS_CIFS', fake_host.storage_protocol)
        self.assertEqual('OpenStack', fake_host.vendor_name)
        self.assertEqual('1.1', fake_host.driver_version)
        self.assertTrue(fake_host.ipv4_support)
        self.assertFalse(fake_host.ipv6_support)

        # Backend level stats remain uninitialized
        self.assertEqual(0, fake_host.total_capacity_gb)
        self.assertIsNone(fake_host.free_capacity_gb)
        # Pool stats has been updated
        self.assertEqual(2, len(fake_host.pools))

        self.assertEqual(500, fake_host.pools['pool1'].total_capacity_gb)
        self.assertEqual(230, fake_host.pools['pool1'].free_capacity_gb)
        self.assertTrue(fake_host.pools['pool1'].ipv4_support)
        self.assertFalse(fake_host.pools['pool1'].ipv6_support)
        self.assertEqual(1024, fake_host.pools['pool2'].total_capacity_gb)
        self.assertEqual(1024, fake_host.pools['pool2'].free_capacity_gb)
        self.assertTrue(fake_host.pools['pool2'].ipv4_support)
        self.assertFalse(fake_host.pools['pool2'].ipv6_support)

        capability = {
            'share_backend_name': 'Backend1',
            'vendor_name': 'OpenStack',
            'driver_version': '1.0',
            'storage_protocol': 'NFS_CIFS',
            'pools': [
                {'pool_name': 'pool3',
                 'total_capacity_gb': 10000,
                 'free_capacity_gb': 10000,
                 'allocated_capacity_gb': 0,
                 'qos': 'False',
                 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 },
            ],
            'timestamp': None,
        }

        # test update HostState Record
        fake_host.update_from_share_capability(capability,
                                               context=fake_context)

        self.assertEqual('1.0', fake_host.driver_version)

        # Non-active pool stats has been removed
        self.assertEqual(1, len(fake_host.pools))

        self.assertRaises(KeyError, lambda: fake_host.pools['pool1'])
        self.assertRaises(KeyError, lambda: fake_host.pools['pool2'])

        self.assertEqual(10000, fake_host.pools['pool3'].total_capacity_gb)
        self.assertEqual(10000, fake_host.pools['pool3'].free_capacity_gb)

    def test_update_from_share_unknown_capability(self):
        share_capability = {
            'total_capacity_gb': 'unknown',
            'free_capacity_gb': 'unknown',
            'allocated_capacity_gb': 1,
            'reserved_percentage': 0,
            'reserved_snapshot_percentage': 0,
            'timestamp': None
        }
        fake_context = context.RequestContext('user', 'project', is_admin=True)
        fake_host = host_manager.HostState('host1#_pool0')
        self.assertIsNone(fake_host.free_capacity_gb)
        fake_host.update_from_share_capability(share_capability,
                                               context=fake_context)
        # Backend level stats remain uninitialized
        self.assertEqual(fake_host.total_capacity_gb, 0)
        self.assertIsNone(fake_host.free_capacity_gb)
        # Pool stats has been updated
        self.assertEqual(fake_host.pools['_pool0'].total_capacity_gb,
                         'unknown')
        self.assertEqual(fake_host.pools['_pool0'].free_capacity_gb,
                         'unknown')

    def test_consume_from_share_capability(self):
        fake_context = context.RequestContext('user', 'project', is_admin=True)
        share_size = 10
        free_capacity = 100
        provisioned_capacity_gb = 50
        fake_share = {'id': 'foo', 'size': share_size}
        share_capability = {
            'total_capacity_gb': free_capacity * 2,
            'free_capacity_gb': free_capacity,
            'provisioned_capacity_gb': provisioned_capacity_gb,
            'allocated_capacity_gb': provisioned_capacity_gb,
            'reserved_percentage': 0,
            'reserved_snapshot_percentage': 0,
            'timestamp': None
        }
        fake_host = host_manager.PoolState('host1', share_capability, '_pool0')

        fake_host.update_from_share_capability(share_capability,
                                               context=fake_context)
        fake_host.consume_from_share(fake_share)
        self.assertEqual(fake_host.free_capacity_gb,
                         free_capacity - share_size)
        self.assertEqual(fake_host.provisioned_capacity_gb,
                         provisioned_capacity_gb + share_size)
        self.assertEqual(fake_host.allocated_capacity_gb,
                         provisioned_capacity_gb + share_size)

    def test_consume_from_share_unknown_capability(self):
        share_capability = {
            'total_capacity_gb': 'unknown',
            'free_capacity_gb': 'unknown',
            'provisioned_capacity_gb': None,
            'allocated_capacity_gb': 0,
            'reserved_percentage': 0,
            'reserved_snapshot_percentage': 0,
            'timestamp': None
        }
        fake_context = context.RequestContext('user', 'project', is_admin=True)
        fake_host = host_manager.PoolState('host1', share_capability, '_pool0')
        share_size = 1000
        fake_share = {'id': 'foo', 'size': share_size}

        fake_host.update_from_share_capability(share_capability,
                                               context=fake_context)
        fake_host.consume_from_share(fake_share)
        self.assertEqual(fake_host.total_capacity_gb, 'unknown')
        self.assertEqual(fake_host.free_capacity_gb, 'unknown')
        self.assertIsNone(fake_host.provisioned_capacity_gb)
        self.assertEqual(fake_host.allocated_capacity_gb, share_size)

    def test_consume_from_share_invalid_capacity(self):
        fake_host = host_manager.PoolState('host1', {}, '_pool0')
        fake_host.free_capacity_gb = 'invalid_foo_string'
        fake_host.provisioned_capacity_gb = None
        fake_host.allocated_capacity_gb = 0
        fake_share = {'id': 'fake', 'size': 10}

        self.assertRaises(exception.InvalidCapacity,
                          fake_host.consume_from_share, fake_share)

    def test_repr(self):

        capability = {
            'share_backend_name': 'Backend1',
            'vendor_name': 'OpenStack',
            'driver_version': '1.0',
            'storage_protocol': 'NFS_CIFS',
            'total_capacity_gb': 20000,
            'free_capacity_gb': 15000,
            'allocated_capacity_gb': 5000,
            'timestamp': None,
            'reserved_percentage': 0,
            'reserved_snapshot_percentage': 0,
        }
        fake_context = context.RequestContext('user', 'project', is_admin=True)
        fake_host = host_manager.HostState('host1')
        fake_host.update_from_share_capability(capability,
                                               context=fake_context)

        result = fake_host.__repr__()
        expected = ("host: 'host1', free_capacity_gb: None, "
                    "pools: {'Backend1': host: 'host1#Backend1', "
                    "free_capacity_gb: 15000, pools: None}")
        self.assertEqual(expected, result)


@ddt.ddt
class PoolStateTestCase(test.TestCase):
    """Test case for HostState class."""

    @ddt.data(
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'thin_provisioning': True,
                 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 4,
                        'updated_at': timeutils.utcnow()
                    },
                    {
                        'id': 2, 'host': 'host1',
                        'status': 'available',
                        'share_id': 12, 'size': None,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'thin_provisioning': False, 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                    {
                        'id': 2, 'host': 'host1',
                        'status': 'available',
                        'share_id': 12, 'size': None,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'thin_provisioning': [False], 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                    {
                        'id': 2, 'host': 'host1',
                        'status': 'available',
                        'share_id': 12, 'size': None,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'thin_provisioning': [True, False], 'cap1': 'val1',
                 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 4,
                        'updated_at': timeutils.utcnow()
                    },
                    {
                        'id': 2, 'host': 'host1',
                        'status': 'available',
                        'share_id': 12, 'size': None,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'cap1': 'val1', 'cap2': 'val2', 'ipv4_support': True,
                 'ipv6_support': False},
            'instances': []
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'thin_provisioning': True, 'cap1': 'val1', 'cap2': 'val2',
                 'ipv4_support': True, 'ipv6_support': False},
            'instances': []
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'thin_provisioning': [False], 'cap1': 'val1', 'cap2': 'val2',
                 'ipv4_support': True, 'ipv6_support': False},
            'instances': []
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'reserved_percentage': 0, 'timestamp': None,
                 'reserved_snapshot_percentage': 0,
                 'thin_provisioning': [True, False], 'cap1': 'val1',
                 'cap2': 'val2', 'ipv4_support': True, 'ipv6_support': False},
            'instances': []
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'allocated_capacity_gb': 256, 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2',
                 'ipv4_support': False, 'ipv6_support': True
                 },
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 4,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'allocated_capacity_gb': 256, 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2',
                 'ipv4_support': True, 'ipv6_support': True},
            'instances': []
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'provisioned_capacity_gb': 256, 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2',
                 'ipv4_support': False, 'ipv6_support': False
                 },
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'allocated_capacity_gb': 256, 'provisioned_capacity_gb': 1,
                 'thin_provisioning': True, 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'allocated_capacity_gb': 256, 'provisioned_capacity_gb': 1,
                 'thin_provisioning': [False], 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'allocated_capacity_gb': 256, 'provisioned_capacity_gb': 1,
                 'thin_provisioning': [True, False], 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'allocated_capacity_gb': 256, 'provisioned_capacity_gb': 256,
                 'thin_provisioning': False, 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
        {
            'share_capability':
                {'total_capacity_gb': 1024, 'free_capacity_gb': 512,
                 'allocated_capacity_gb': 256, 'provisioned_capacity_gb': 256,
                 'thin_provisioning': [False], 'reserved_percentage': 0,
                 'reserved_snapshot_percentage': 0,
                 'timestamp': None, 'cap1': 'val1', 'cap2': 'val2'},
            'instances':
                [
                    {
                        'id': 1, 'host': 'host1',
                        'status': 'available',
                        'share_id': 11, 'size': 1,
                        'updated_at': timeutils.utcnow()
                    },
                ]
        },
    )
    @ddt.unpack
    def test_update_from_share_capability(self, share_capability, instances):
        fake_context = context.RequestContext('user', 'project', is_admin=True)
        self.mock_object(
            db, 'share_instances_get_all_by_host',
            mock.Mock(return_value=instances))
        fake_pool = host_manager.PoolState('host1', None, 'pool0')
        self.assertIsNone(fake_pool.free_capacity_gb)

        fake_pool.update_from_share_capability(share_capability,
                                               context=fake_context)

        self.assertEqual('host1#pool0', fake_pool.host)
        self.assertEqual('pool0', fake_pool.pool_name)
        self.assertEqual(1024, fake_pool.total_capacity_gb)
        self.assertEqual(512, fake_pool.free_capacity_gb)
        self.assertDictEqual(share_capability, dict(fake_pool.capabilities))

        if 'thin_provisioning' in share_capability:
            thin_provisioned = scheduler_utils.thin_provisioning(
                share_capability['thin_provisioning'])
        else:
            thin_provisioned = False

        if thin_provisioned:
            self.assertEqual(thin_provisioned, fake_pool.thin_provisioning)
            if 'provisioned_capacity_gb' not in share_capability or (
                    share_capability['provisioned_capacity_gb'] is None):
                db.share_instances_get_all_by_host.assert_called_once_with(
                    fake_context, fake_pool.host, with_share_data=True)
                if len(instances) > 0:
                    self.assertEqual(4, fake_pool.provisioned_capacity_gb)
                else:
                    self.assertEqual(0, fake_pool.provisioned_capacity_gb)
            else:
                self.assertFalse(db.share_instances_get_all_by_host.called)
                self.assertEqual(share_capability['provisioned_capacity_gb'],
                                 fake_pool.provisioned_capacity_gb)
        else:
            self.assertFalse(fake_pool.thin_provisioning)
            self.assertFalse(db.share_instances_get_all_by_host.called)
            if 'provisioned_capacity_gb' not in share_capability or (
                    share_capability['provisioned_capacity_gb'] is None):
                self.assertIsNone(fake_pool.provisioned_capacity_gb)
            else:
                self.assertEqual(share_capability['provisioned_capacity_gb'],
                                 fake_pool.provisioned_capacity_gb)

        if 'allocated_capacity_gb' in share_capability:
            self.assertEqual(share_capability['allocated_capacity_gb'],
                             fake_pool.allocated_capacity_gb)
        else:
            self.assertEqual(0, fake_pool.allocated_capacity_gb)

        if 'ipv4_support' in share_capability:
            self.assertEqual(share_capability['ipv4_support'],
                             fake_pool.ipv4_support)
        if 'ipv6_support' in share_capability:
            self.assertEqual(share_capability['ipv6_support'],
                             fake_pool.ipv6_support)
